from fastapi import FastAPI

from viper_hrunner.vhrun import __version__
from .routers import deps, debugtalk, debug

app = FastAPI()


@app.get("/hrun/version")
async def get_hrun_version():
    return {"code": 0, "message": "success", "result": {"HttpRunner": __version__}}


app.include_router(deps.router)
app.include_router(debugtalk.router)
app.include_router(debug.router)



__version__ = "3.1.6"
__description__ = "One-stop solution for HTTP(S) testing."

# import firstly for monkey patch if needed
from viper_hrunner.vhrun.ext.locust import main_locusts
from viper_hrunner.vhrun.parser import parse_parameters as Parameters
from viper_hrunner.vhrun.runner import HttpRunner
from viper_hrunner.vhrun.testcase import Config, Step, RunRequest, RunTestCase

__all__ = [
    "__version__",
    "__description__",
    "HttpRunner",
    "Config",
    "Step",
    "RunRequest",
    "RunTestCase",
    "Parameters",
]



""" Convert HAR (HTTP Archive) to YAML/JSON testcase for HttpRunner.

Usage:
    # convert to JSON format testcase
    $ hrun har2case demo.har

    # convert to YAML format testcase
    $ hrun har2case demo.har -2y

"""

from viper_hrunner.vhrun.ext.har2case.core import HarParser
from sentry_sdk import capture_message


def init_har2case_parser(subparsers):
    """ HAR converter: parse command line options and run commands.
    """
    parser = subparsers.add_parser(
        "har2case",
        help="Convert HAR(HTTP Archive) to YAML/JSON testcases for HttpRunner.",
    )
    parser.add_argument("har_source_file", nargs="?", help="Specify HAR source file")
    parser.add_argument(
        "-2y",
        "--to-yml",
        "--to-yaml",
        dest="to_yaml",
        action="store_true",
        help="Convert to YAML format, if not specified, convert to pytest format by default.",
    )
    parser.add_argument(
        "-2j",
        "--to-json",
        dest="to_json",
        action="store_true",
        help="Convert to JSON format, if not specified, convert to pytest format by default.",
    )
    parser.add_argument(
        "--filter",
        help="Specify filter keyword, only url include filter string will be converted.",
    )
    parser.add_argument(
        "--exclude",
        help="Specify exclude keyword, url that includes exclude string will be ignored, "
        "multiple keywords can be joined with '|'",
    )

    return parser


def main_har2case(args):
    har_source_file = args.har_source_file

    if args.to_yaml:
        output_file_type = "YAML"
    elif args.to_json:
        output_file_type = "JSON"
    else:
        output_file_type = "pytest"

    capture_message(f"har2case {output_file_type}")
    HarParser(har_source_file, args.filter, args.exclude).gen_testcase(output_file_type)

    return 0



import io
import json
import os
from datetime import datetime
from jinja2 import Template
from loguru import logger


def gen_html_report(summary, report_template=None, report_dir=None, report_name=None) -> str:
    """ render html report with specified report name and template

    Args:
        summary (dict): test result summary data
        report_template (str): specify html report template path, template should be in Jinja2 format.
        report_dir (str): specify html report save directory
        report_name (str): report name.

    """
    # if not summary["time"] or summary["stat"]["testcases"]["total"] == 0:
    #     logger.log_error("test result summary is empty ! {}".format(summary))
    #     raise SummaryEmpty

    if not report_template:
        report_template = os.path.join(
            os.path.abspath(os.path.dirname(__file__)),
            "template.html"
        )
        # logger.debug("No html report template specified, use default.")
    else:
        logger.info("render with html report template: {}".format(report_template))

    logger.info("1, Start to render Html report ...")

    start_at_timestamp = summary["time"]["start_at"]
    # utc_time_iso_8601_str = datetime.utcfromtimestamp(start_at_timestamp).isoformat()
    utc_time_iso_8601_str = datetime.fromtimestamp(start_at_timestamp).strftime('%Y_%m_%d_%H_%M_%S')
    utc_time_iso_8601_str = os.environ["test_start_time"] if "test_start_time" in os.environ else utc_time_iso_8601_str
    summary["time"]["start_datetime"] = datetime.fromtimestamp(start_at_timestamp).strftime('%Y/%m/%d %H:%M:%S')

    report_dir = report_dir or os.path.join(os.getcwd(), "reports")
    if report_name:
        report_file_name = report_name
    else:
        # fix #826: Windows does not support file name include ":"
        report_file_name = "{}.html".format(utc_time_iso_8601_str.replace(":", "").replace("-", ""))

    if not os.path.isdir(report_dir):
        os.makedirs(report_dir)

    report_path = os.path.join(report_dir, report_file_name)
    with io.open(report_template, "r", encoding='utf-8') as fp_r:
        template_content = fp_r.read()
        with io.open(report_path, 'w', encoding='utf-8') as fp_w:
            rendered_content = Template(
                template_content,
                extensions=["jinja2.ext.loopcontrols"]
            ).render(summary)
            fp_w.write(rendered_content)

    logger.info("1, Generated Html report: {}".format(report_path))
    os.environ["report_path"] = report_path

    json_file_name = "{}.json".format(utc_time_iso_8601_str.replace(":", "").replace("-", ""))
    json_file_path = os.path.join(report_dir, json_file_name)
    # with open(json_file_path, "w") as f:
    #     f.write(json.dumps(summary))

    return report_path




import copy
import json
from typing import Dict, Text, Any, NoReturn

import jmespath
import requests
from jmespath.exceptions import JMESPathError
from loguru import logger

from viper_hrunner.vhrun import exceptions
from viper_hrunner.vhrun.exceptions import ValidationFailure, ParamsError
from viper_hrunner.vhrun.models import VariablesMapping, Validators, FunctionsMapping
from viper_hrunner.vhrun.parser import parse_data, parse_string_value, get_mapping_function

try:
    JSONDecodeError = json.JSONDecodeError
except AttributeError:
    JSONDecodeError = ValueError


def get_uniform_comparator(comparator: Text):
    """ convert comparator alias to uniform name
    """
    if comparator in ["eq", "equals", "equal"]:
        return "equal"
    elif comparator in ["lt", "less_than"]:
        return "less_than"
    elif comparator in ["le", "less_or_equals"]:
        return "less_or_equals"
    elif comparator in ["gt", "greater_than"]:
        return "greater_than"
    elif comparator in ["ge", "greater_or_equals"]:
        return "greater_or_equals"
    elif comparator in ["ne", "not_equal"]:
        return "not_equal"
    elif comparator in ["str_eq", "string_equals"]:
        return "string_equals"
    elif comparator in ["len_eq", "length_equal"]:
        return "length_equal"
    elif comparator in [
        "len_gt",
        "length_greater_than",
    ]:
        return "length_greater_than"
    elif comparator in [
        "len_ge",
        "length_greater_or_equals",
    ]:
        return "length_greater_or_equals"
    elif comparator in ["len_lt", "length_less_than"]:
        return "length_less_than"
    elif comparator in [
        "len_le",
        "length_less_or_equals",
    ]:
        return "length_less_or_equals"
    else:
        return comparator


def uniform_validator(validator):
    """ unify validator

    Args:
        validator (dict): validator maybe in two formats:

            format1: this is kept for compatibility with the previous versions.
                {"check": "status_code", "comparator": "eq", "expect": 201}
                {"check": "$resp_body_success", "comparator": "eq", "expect": True}
            format2: recommended new version, {assert: [check_item, expected_value]}
                {'eq': ['status_code', 201]}
                {'eq': ['$resp_body_success', True]}

    Returns
        dict: validator info

            {
                "check": "status_code",
                "expect": 201,
                "assert": "equals"
            }

    """
    if not isinstance(validator, dict):
        raise ParamsError(f"invalid validator: {validator}")

    if "check" in validator and "expect" in validator:
        # format1
        check_item = validator["check"]
        expect_value = validator["expect"]
        message = validator.get("message", "")
        comparator = validator.get("comparator", "eq")

    elif len(validator) == 1:
        # format2
        comparator = list(validator.keys())[0]
        compare_values = validator[comparator]

        if not isinstance(compare_values, list) or len(compare_values) not in [2, 3]:
            raise ParamsError(f"invalid validator: {validator}")

        check_item = compare_values[0]
        expect_value = compare_values[1]
        if len(compare_values) == 3:
            message = compare_values[2]
        else:
            # len(compare_values) == 2
            message = ""

    else:
        raise ParamsError(f"invalid validator: {validator}")

    # uniform comparator, e.g. lt => less_than, eq => equals
    assert_method = get_uniform_comparator(comparator)

    return {
        "check": check_item,
        "expect": expect_value,
        "assert": assert_method,
        "message": message,
    }


class ResponseObject(object):
    def __init__(self, resp_obj):
        """ initialize with a requests.Response object

        Args:
            resp_obj (instance): requests.Response instance

        """
        self.resp_obj = resp_obj
        self.validation_results: Dict = {}

    def __getattr__(self, key):
        if key in ["json", "content", "body"]:
            try:
                value = self.resp_obj.json()
            except ValueError:
                value = self.resp_obj.content
        elif key == "cookies":
            value = self.resp_obj.cookies.get_dict()
        else:
            try:
                value = getattr(self.resp_obj, key)
            except AttributeError:
                err_msg = "ResponseObject does not have attribute: {}".format(key)
                logger.error(err_msg)
                raise exceptions.ParamsError(err_msg)

        self.__dict__[key] = value
        return value

    def _search_jmespath(self, expr: Text) -> Any:
        if isinstance(self.resp_obj, requests.Response):
            resp_obj_meta = {
                "status_code": self.status_code,
                "headers": self.headers,
                "cookies": self.cookies,
                "body": self.body,
            }
        else:
            resp_obj_meta = {
                "request": self.resp_obj.get("request"),
                "body": self.resp_obj.get("body")
            }
        if not expr.startswith(tuple(resp_obj_meta.keys())):
            return expr
        
        try:
            check_value = jmespath.search(expr, resp_obj_meta)
        except JMESPathError as ex:
            logger.error(
                f"failed to search with jmespath\n"
                f"expression: {expr}\n"
                f"data: {resp_obj_meta}\n"
                f"exception: {ex}"
            )
            raise

        return check_value

    def extract(self, extractors: Dict[Text, Text]) -> Dict[Text, Any]:
        if not extractors:
            return {}

        extract_mapping = {}
        for key, field in extractors.items():
            field_value = self._search_jmespath(field)
            extract_mapping[key] = field_value

        logger.info(f"extract mapping: {extract_mapping}")
        return extract_mapping

    def validate(
        self,
        validators: Validators,
        variables_mapping: VariablesMapping = None,
        functions_mapping: FunctionsMapping = None,
    ) -> NoReturn:

        variables_mapping = variables_mapping or {}
        functions_mapping = functions_mapping or {}

        validate_list = []
        self.validation_results = {}

        if not validators:
            return validate_list, ""

        validate_pass = True
        failures = []

        for v in validators:

            if "validate_extractor" not in self.validation_results:
                self.validation_results["validate_extractor"] = []

            u_validator = uniform_validator(v)

            # check item
            check_item = u_validator["check"]
            if "$" in check_item:
                # check_item is variable or function
                check_item = parse_data(
                    check_item, variables_mapping, functions_mapping
                )
                check_item = parse_string_value(check_item)

            if check_item and isinstance(check_item, Text):
                check_value = self._search_jmespath(check_item)
            else:
                # variable or function evaluation result is "" or not text
                check_value = check_item

            # comparator
            assert_method = u_validator["assert"]
            assert_func = get_mapping_function(assert_method, functions_mapping)

            # expect item
            expect_item = u_validator["expect"]
            # parse expected value with config/teststep/extracted variables
            expect_value = parse_data(expect_item, variables_mapping, functions_mapping)

            # message
            message = u_validator["message"]
            # parse message with config/teststep/extracted variables
            message = parse_data(message, variables_mapping, functions_mapping)

            validate_msg = f"assert {check_item} {assert_method} {expect_value}({type(expect_value).__name__})"

            validator_dict = {
                "comparator": assert_method,
                "check": check_item,
                "check_value": check_value,
                "expect": expect_item,
                "expect_value": expect_value,
                "message": message,
            }

            valiate_desc = copy.deepcopy(validate_msg)
            try:
                assert_func(check_value, expect_value, message)
                validate_msg += "\t==> pass"
                logger.info(validate_msg)
                validator_dict["check_result"] = "pass"
            except AssertionError as ex:
                validate_pass = False
                validator_dict["check_result"] = "fail"
                validate_msg = "\n\n" + validate_msg
                validate_msg += "\t==> fail"
                validate_msg += (
                    f"\n"
                    f"check_item: {check_item}\n"
                    f"check_value: {check_value}({type(check_value).__name__})\n"
                    f"assert_method: {assert_method}\n"
                    f"expect_value: {expect_value}({type(expect_value).__name__})"
                )
                message = str(ex)
                if message:
                    validate_msg += f"\nmessage: {message}"

                logger.error(validate_msg)
                failures.append(validate_msg)

            self.validation_results["validate_extractor"].append(validator_dict)

            test_result = validator_dict["check_result"].upper()
            validate_list.append(("{} -> {}".format(test_result, valiate_desc), validator_dict, test_result))

        if not validate_pass:
            failures_string = "\n\n".join([failure for failure in failures])
            return validate_list, failures_string
        else:
            return validate_list, ""


def query_json(json_content, query, delimiter='.'):
    raise_flag = False
    response_body = u"response body: {}\n".format(json_content)
    try:
        for key in query.split(delimiter):
            if isinstance(json_content, (list, str)):
                json_content = json_content[int(key)]
            elif isinstance(json_content, dict):
                json_content = json_content[key]
            else:
                logger.error(
                    "invalid type value: {}({})".format(json_content, type(json_content)))
                raise_flag = True
    except (KeyError, ValueError, IndexError):
        raise_flag = True

    if raise_flag:
        err_msg = u"Failed to extract! => {}\n".format(query)
        err_msg += response_body
        logger.error(err_msg)
        raise exceptions.ExtractFailure(err_msg)

    return json_content


class SSHResponseObject(ResponseObject):
    def _extract_field_with_delimiter(self, field):
        # string.split(sep=None, maxsplit=1) -> list of strings
        # e.g. "content.person.name" => ["content", "person.name"]
        try:
            top_query, sub_query = field.split('.', 1)
        except ValueError:
            top_query = field
            sub_query = None

        # response body
        if top_query in ["body", "content", "text", "json"]:
            try:
                body = self.resp_obj.content
            except JSONDecodeError:
                body = self.text

            if not sub_query:
                # extract response body
                return body

            if isinstance(body, (dict, list)):
                # content = {"xxx": 123}, content.xxx
                return query_json(body, sub_query)
            elif sub_query.isdigit():
                # content = "abcdefg", content.3 => d
                return query_json(body, sub_query)
            else:
                # content = "<html>abcdefg</html>", content.xxx
                err_msg = u"Failed to extract attribute from response body! => {}\n".format(field)
                err_msg += u"response body: {}\n".format(body)
                logger.error(err_msg)
                raise exceptions.ExtractFailure(err_msg)

        # new set response attributes in teardown_hooks
        elif top_query in self.__dict__:
            attributes = self.__dict__[top_query]

            if not sub_query:
                # extract response attributes
                return attributes

            if isinstance(attributes, (dict, list)):
                # attributes = {"xxx": 123}, content.xxx
                return query_json(attributes, sub_query)
            elif sub_query.isdigit():
                # attributes = "abcdefg", attributes.3 => d
                return query_json(attributes, sub_query)
            else:
                # content = "attributes.new_attribute_not_exist"
                err_msg = u"Failed to extract cumstom set attribute from teardown hooks! => {}\n".format(field)
                err_msg += u"response set attributes: {}\n".format(attributes)
                logger.error(err_msg)
                raise exceptions.TeardownHooksFailure(err_msg)

        # others
        else:
            err_msg = u"Failed to extract attribute from response! => {}\n".format(field)
            err_msg += u"available response attributes: status_code, cookies, elapsed, headers, content, " \
                       u"text, json, encoding, ok, reason, url.\n\n"
            err_msg += u"If you want to set attribute in teardown_hooks, take the following example as reference:\n"
            err_msg += u"response.new_attribute = 'new_attribute_value'\n"
            logger.error(err_msg)
            raise exceptions.ParamsError(err_msg)



import random

from locust import task, HttpUser, between

from viper_hrunner.vhrun.ext.locust import prepare_locust_tests


class HttpRunnerUser(HttpUser):
    host = ""
    wait_time = between(5, 15)

    def on_start(self):
        locust_tests = prepare_locust_tests()
        self.testcase_runners = [
            testcase().with_session(self.client) for testcase in locust_tests
        ]

    @task
    def test_any(self):
        test_runner = random.choice(self.testcase_runners)
        try:
            test_runner.run()
        except Exception as ex:
            self.environment.events.request_failure.fire(
                request_type="Failed",
                name=test_runner.config.name,
                response_time=0,
                response_length=0,
                exception=ex,
            )



"""
Built-in validate comparators.
"""

import re
from typing import Text, Any, Union


def equal(check_value: Any, expect_value: Any, message: Text = ""):
    assert check_value == expect_value, message


def greater_than(
    check_value: Union[int, float], expect_value: Union[int, float], message: Text = ""
):
    assert check_value > expect_value, message


def less_than(
    check_value: Union[int, float], expect_value: Union[int, float], message: Text = ""
):
    assert check_value < expect_value, message


def greater_or_equals(
    check_value: Union[int, float], expect_value: Union[int, float], message: Text = ""
):
    assert check_value >= expect_value, message


def less_or_equals(
    check_value: Union[int, float], expect_value: Union[int, float], message: Text = ""
):
    assert check_value <= expect_value, message


def not_equal(check_value: Any, expect_value: Any, message: Text = ""):
    assert check_value != expect_value, message


def string_equals(check_value: Text, expect_value: Any, message: Text = ""):
    assert str(check_value) == str(expect_value), message


def length_equal(check_value: Text, expect_value: int, message: Text = ""):
    assert isinstance(expect_value, int), "expect_value should be int type"
    assert len(check_value) == expect_value, message


def length_greater_than(
    check_value: Text, expect_value: Union[int, float], message: Text = ""
):
    assert isinstance(
        expect_value, (int, float)
    ), "expect_value should be int/float type"
    assert len(check_value) > expect_value, message


def length_greater_or_equals(
    check_value: Text, expect_value: Union[int, float], message: Text = ""
):
    assert isinstance(
        expect_value, (int, float)
    ), "expect_value should be int/float type"
    assert len(check_value) >= expect_value, message


def length_less_than(
    check_value: Text, expect_value: Union[int, float], message: Text = ""
):
    assert isinstance(
        expect_value, (int, float)
    ), "expect_value should be int/float type"
    assert len(check_value) < expect_value, message


def length_less_or_equals(
    check_value: Text, expect_value: Union[int, float], message: Text = ""
):
    assert isinstance(
        expect_value, (int, float)
    ), "expect_value should be int/float type"
    assert len(check_value) <= expect_value, message


def contains(check_value: Any, expect_value: Any, message: Text = ""):
    assert isinstance(
        check_value, (list, tuple, dict, str, bytes)
    ), "expect_value should be list/tuple/dict/str/bytes type"
    assert expect_value in check_value, message


def contained_by(check_value: Any, expect_value: Any, message: Text = ""):
    assert isinstance(
        expect_value, (list, tuple, dict, str, bytes)
    ), "expect_value should be list/tuple/dict/str/bytes type"
    assert check_value in expect_value, message


def type_match(check_value: Any, expect_value: Any, message: Text = ""):
    def get_type(name):
        if isinstance(name, type):
            return name
        elif isinstance(name, str):
            try:
                return __builtins__[name]
            except KeyError:
                raise ValueError(name)
        else:
            raise ValueError(name)

    if expect_value in ["None", "NoneType", None]:
        assert check_value is None, message
    else:
        assert type(check_value) == get_type(expect_value), message


def regex_match(check_value: Text, expect_value: Any, message: Text = ""):
    assert isinstance(expect_value, str), "expect_value should be Text type"
    assert isinstance(check_value, str), "check_value should be Text type"
    assert re.match(expect_value, check_value), message


def startswith(check_value: Any, expect_value: Any, message: Text = ""):
    assert str(check_value).startswith(str(expect_value)), message


def endswith(check_value: Text, expect_value: Any, message: Text = ""):
    assert str(check_value).endswith(str(expect_value)), message



import allure


def teststep_allure_step(step_desc, request_dict, response_dict, validate_list, extract_mapping):
    @allure.step
    def request_data(**kv):
        pass

    @allure.step
    def response_data(**kv):
        pass

    def validate_allure_step(valiate_desc, validator_dict):
        @allure.step
        def validate_detail(**kv):
            pass

        @allure.step(valiate_desc)
        def assert_step(test_result):
            validate_detail(**validator_dict)

        return assert_step

    @allure.step
    def extract_values(**kv):
        pass

    @allure.step(step_desc)
    def step_function():
        request_data(**request_dict)
        response_data(**response_dict)
        with allure.step("validate_list"):
            for desc, detail, test_result in validate_list:
                if "{" in desc:
                    desc = desc.replace("{", "<")
                    desc = desc.replace("}", ">")
                validate_allure_step(desc, detail)(test_result)
        extract_values(**extract_mapping)

    step_function()




import sys

if "locust" in sys.argv[0]:
    try:
        # monkey patch all at beginning to avoid RecursionError when running locust.
        # `from gevent import monkey; monkey.patch_all()` will be triggered when importing locust
        from locust import main as locust_main

        print("NOTICE: gevent monkey patches have been applied !!!")
    except ImportError:
        msg = """
Locust is not installed, install first and try again.
install with pip:
$ pip install locust
"""
        print(msg)
        sys.exit(1)

import importlib.util
import inspect
import os
from typing import List

from loguru import logger


""" converted pytest files from YAML/JSON testcases
"""
pytest_files: List = []


def is_httprunner_testcase(item):
    """ check if a variable is a HttpRunner testcase class
    """
    from viper_hrunner.vhrun import HttpRunner

    # TODO: skip referenced testcase
    return bool(
        inspect.isclass(item)
        and issubclass(item, HttpRunner)
        and item.__name__ != "HttpRunner"
    )


def prepare_locust_tests() -> List:
    """ prepare locust testcases

    Returns:
        list: testcase class list
    """

    locust_tests = []

    for pytest_file in pytest_files:
        spec = importlib.util.spec_from_file_location("module.name", pytest_file)
        module = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(module)

        for name, item in vars(module).items():

            if not is_httprunner_testcase(item):
                continue

            for _ in range(item.config.weight):
                locust_tests.append(item)

    return locust_tests


def main_locusts():
    """ locusts entrance
    """
    from viper_hrunner.vhrun.utils import init_sentry_sdk
    from sentry_sdk import capture_message

    init_sentry_sdk()
    capture_message("start to run locusts")

    # avoid print too much log details in console
    # logger.remove()
    # logger.add(sys.stderr, level="WARNING")

    sys.argv[0] = "locust"
    if len(sys.argv) == 1:
        sys.argv.extend(["-h"])

    if sys.argv[1] in ["-h", "--help", "-V", "--version"]:
        locust_main.main()

    def get_arg_index(*target_args):
        for arg in target_args:
            if arg not in sys.argv:
                continue

            return sys.argv.index(arg) + 1

        return None

    # get testcase file path
    testcase_index = get_arg_index("-f", "--locustfile")
    if not testcase_index:
        print("Testcase file is not specified, exit 1.")
        sys.exit(1)

    from viper_hrunner.vhrun.make import main_make

    global pytest_files
    testcase_file_path = sys.argv[testcase_index]
    pytest_files = main_make([testcase_file_path])
    if not pytest_files:
        print("No valid testcases found, exit 1.")
        sys.exit(1)

    sys.argv[testcase_index] = os.path.join(os.path.dirname(__file__), "locustfile.py")

    locust_main.main()



""" upload test extension.

If you want to use this extension, you should install the following dependencies first.

- requests_toolbelt
- filetype

Then you can write upload test script as below:

    - test:
        name: upload file
        request:
            url: http://httpbin.org/upload
            method: POST
            headers:
                Cookie: session=AAA-BBB-CCC
            upload:
                file: "data/file_to_upload"
                field1: "value1"
                field2: "value2"
        validate:
            - eq: ["status_code", 200]

For compatibility, you can also write upload test script in old way:

    - test:
        name: upload file
        variables:
            file: "data/file_to_upload"
            field1: "value1"
            field2: "value2"
            m_encoder: ${multipart_encoder(file=$file, field1=$field1, field2=$field2)}
        request:
            url: http://httpbin.org/upload
            method: POST
            headers:
                Content-Type: ${multipart_content_type($m_encoder)}
                Cookie: session=AAA-BBB-CCC
            data: $m_encoder
        validate:
            - eq: ["status_code", 200]

"""

import os
import sys
from typing import Text, NoReturn

from viper_hrunner.vhrun.models import TStep, FunctionsMapping
from viper_hrunner.vhrun.parser import parse_variables_mapping
from loguru import logger

try:
    import filetype
    from requests_toolbelt import MultipartEncoder

    UPLOAD_READY = True
except ModuleNotFoundError:
    UPLOAD_READY = False


def ensure_upload_ready():
    if UPLOAD_READY:
        return

    msg = """
    uploader extension dependencies uninstalled, install first and try again.
    install with pip:
    $ pip install requests_toolbelt filetype

    or you can install viper_hrunner.vhrun with optional upload dependencies:
    $ pip install "viper_hrunner.vhrun[upload]"
    """
    logger.error(msg)
    sys.exit(1)


def prepare_upload_step(step: TStep, functions: FunctionsMapping) -> "NoReturn":
    """ preprocess for upload test
        replace `upload` info with MultipartEncoder

    Args:
        step: teststep
            {
                "variables": {},
                "request": {
                    "url": "http://httpbin.org/upload",
                    "method": "POST",
                    "headers": {
                        "Cookie": "session=AAA-BBB-CCC"
                    },
                    "upload": {
                        "file": "data/file_to_upload"
                        "md5": "123"
                    }
                }
            }
        functions: functions mapping

    """
    print(dir(step.request))
    if not step.request.upload:
        return

    ensure_upload_ready()
    params_list = []
    for key, value in step.request.upload.items():
        step.variables[key] = value
        params_list.append(f"{key}=${key}")

    params_str = ", ".join(params_list)
    step.variables["m_encoder"] = "${multipart_encoder(" + params_str + ")}"

    # parse variables
    step.variables = parse_variables_mapping(step.variables, functions)

    step.request.headers["Content-Type"] = "${multipart_content_type($m_encoder)}"

    step.request.data = "$m_encoder"


def multipart_encoder(**kwargs):
    """ initialize MultipartEncoder with uploading fields.

    Returns:
        MultipartEncoder: initialized MultipartEncoder object

    """

    def get_filetype(file_path):
        file_type = filetype.guess(file_path)
        if file_type:
            return file_type.mime
        else:
            return "text/html"

    ensure_upload_ready()
    fields_dict = {}
    for key, value in kwargs.items():

        if os.path.isabs(value):
            # value is absolute file path
            _file_path = value
            is_exists_file = os.path.isfile(value)
        else:
            # value is not absolute file path, check if it is relative file path
            from viper_hrunner.vhrun.loader import load_project_meta

            project_meta = load_project_meta("")

            _file_path = os.path.join(project_meta.RootDir, value)
            is_exists_file = os.path.isfile(_file_path)

        if is_exists_file:
            # value is file path to upload
            filename = os.path.basename(_file_path)
            mime_type = get_filetype(_file_path)
            # TODO: fix ResourceWarning for unclosed file
            file_handler = open(_file_path, "rb")
            fields_dict[key] = (filename, file_handler, mime_type)
        else:
            fields_dict[key] = value

    return MultipartEncoder(fields=fields_dict)


def multipart_content_type(m_encoder) -> Text:
    """ prepare Content-Type for request headers

    Args:
        m_encoder: MultipartEncoder object

    Returns:
        content type

    """
    ensure_upload_ready()
    return m_encoder.content_type



import os
from enum import Enum
from typing import Any
from typing import Dict, Text, Union, Callable
from typing import List

from pydantic import BaseModel, Field
from pydantic import HttpUrl

Name = Text
Url = Text
BaseUrl = Union[HttpUrl, Text]
VariablesMapping = Dict[Text, Any]
FunctionsMapping = Dict[Text, Callable]
Headers = Dict[Text, Text]
Cookies = Dict[Text, Text]
Verify = bool
Hooks = List[Union[Text, Dict[Text, Text]]]
Export = List[Text]
Validators = List[Dict]
Env = Dict[Text, Any]


class MethodEnum(Text, Enum):
    GET = "GET"
    POST = "POST"
    PUT = "PUT"
    DELETE = "DELETE"
    HEAD = "HEAD"
    OPTIONS = "OPTIONS"
    PATCH = "PATCH"


class TConfig(BaseModel):
    name: Name
    verify: Verify = False
    base_url: BaseUrl = ""
    # Text: prepare variables in debugtalk.py, ${gen_variables()}
    variables: Union[VariablesMapping, Text] = {}
    parameters: Union[VariablesMapping, Text] = {}
    skip_if_list = []
    setup_hooks: Hooks = []
    teardown_hooks: Hooks = []
    export: Export = []
    path: Text = None
    weight: int = 1


class TRequest(BaseModel):
    """requests.Request model"""

    method: Union[MethodEnum, None] = None
    url: Union[Text, None] = None
    params: Union[Dict[Text, Text], List, Text, None] = {}
    headers: Headers = {}
    req_json: Union[Dict, List, Text] = Field(None, alias="json")
    data: Union[Text, Dict[Text, Any]] = None
    cookies: Cookies = {}
    timeout: float = 120
    allow_redirects: bool = True
    verify: Verify = False
    upload: Dict = {}  # used for upload files


class TStep(BaseModel):
    name: Name
    protocol: Union[Text, None] = None
    ssh_type: Union[Text, None] = None
    connection: Union[Dict, None] = {}
    paths: List[Text] = []
    ssh_cmd: Union[Text, None] = None
    request: Union[TRequest, None] = None
    testcase: Union[Text, Callable, None] = None
    variables: VariablesMapping = {}
    skip_if_list: Hooks = []
    setup_hooks: Hooks = []
    teardown_hooks: Hooks = []
    # used to extract request's response field
    extract: VariablesMapping = {}
    # used to export session variables from referenced testcase
    export: Export = []
    validators: Validators = Field([], alias="validate")
    validate_script: List[Text] = []
    loop_while: Dict = {}
    loop_for: Dict = {}

class TestCase(BaseModel):
    config: TConfig
    teststeps: List[TStep]


class ProjectMeta(BaseModel):
    debugtalk_py: Text = ""  # debugtalk.py file content
    debugtalk_path: Text = ""  # debugtalk.py file path
    dot_env_path: Text = ""  # .env file path
    functions: FunctionsMapping = {}  # functions defined in debugtalk.py
    env: Env = {}
    RootDir: Text = os.getcwd()  # project root directory (ensure absolute), the path debugtalk.py located


class TestsMapping(BaseModel):
    project_meta: ProjectMeta
    testcases: List[TestCase]


class TestCaseTime(BaseModel):
    start_at: float = 0
    start_at_iso_format: Text = ""
    duration: float = 0


class TestCaseInOut(BaseModel):
    config_vars: VariablesMapping = {}
    export_vars: Dict = {}


class RequestStat(BaseModel):
    content_size: float = 0
    response_time_ms: float = 0
    elapsed_ms: float = 0


class AddressData(BaseModel):
    client_ip: Text = "N/A"
    client_port: int = 0
    server_ip: Text = "N/A"
    server_port: int = 0


class RequestData(BaseModel):
    method: MethodEnum = MethodEnum.GET
    url: Url
    headers: Headers = {}
    cookies: Cookies = {}
    body: Union[Text, bytes, List, Dict, None] = {}


class SSHRequestData(BaseModel):
    body: Dict = {}  # for ssh


class ResponseData(BaseModel):
    status_code: int
    headers: Dict
    cookies: Cookies
    encoding: Union[Text, None] = None
    content_type: Text
    body: Union[Text, bytes, List, Dict]


class SSHResponseData(BaseModel):
    body: Union[Text, bytes, List, Dict]


class ReqRespData(BaseModel):
    request: RequestData
    response: ResponseData


class SSHReqRespData(BaseModel):
    request: SSHRequestData
    response: SSHResponseData


class SessionData(BaseModel):
    """request session data, including request, response, validators and stat data"""

    success: bool = False
    # in most cases, req_resps only contains one request & response
    # while when 30X redirect occurs, req_resps will contain multiple request & response
    req_resps: Union[List['ReqRespData'], List['SSHReqRespData']] = []
    stat: RequestStat = RequestStat()
    address: AddressData = AddressData()
    validators: Dict = {}
    err_list: List['Text'] = []
    attachment: Text = ''


class StepData(BaseModel):
    """teststep data, each step maybe corresponding to one request or one testcase"""

    success: bool = False
    name: Text = ""  # teststep name
    data: Union[SessionData, List['StepData']] = None
    export_vars: VariablesMapping = {}

        
StepData.update_forward_refs()


class TestCaseSummary(BaseModel):
    name: Text
    success: bool
    case_id: Text
    time: TestCaseTime
    in_out: TestCaseInOut = {}
    log: Text = ""
    step_datas: List[StepData] = []


class PlatformInfo(BaseModel):
    httprunner_version: Text
    python_version: Text
    platform: Text


class TestCaseRef(BaseModel):
    name: Text
    base_url: Text = ""
    testcase: Text
    variables: VariablesMapping = {}


class TestSuite(BaseModel):
    config: TConfig
    testcases: List[TestCaseRef]


class Stat(BaseModel):
    total: int = 0
    success: int = 0
    fail: int = 0


class TestSuiteSummary(BaseModel):
    success: bool = False
    stat: Stat = Stat()
    time: TestCaseTime = TestCaseTime()
    platform: PlatformInfo
    testcases: List[TestCaseSummary]



import json
import os
import time
import uuid
from datetime import datetime
from typing import List, Dict, Text, NoReturn
from viper_hrunner.vhrun.builtin.functions import get_process_thread_id
import pytest

from viper_hrunner.vhrun.utils import print_info
from viper_hrunner.vhrun.models import SSHRequestData, SSHResponseData
from viper_hrunner.vhrun.models import SessionData, SSHReqRespData
import copy
import traceback

try:
    import allure
    from viper_hrunner.vhrun.report.allure_report.allure_step import teststep_allure_step

    USE_ALLURE = True
except ModuleNotFoundError:
    USE_ALLURE = False

from loguru import logger
from viper_hrunner.vhrun.report.aggregate import aggregate
from viper_hrunner.vhrun import utils, exceptions
from viper_hrunner.vhrun.client import HttpSession
from viper_hrunner.vhrun.exceptions import ValidationFailure, ParamsError
from viper_hrunner.vhrun.ext.uploader import prepare_upload_step
from viper_hrunner.vhrun.loader import load_project_meta, load_testcase_file
from viper_hrunner.vhrun.parser import build_url, parse_data, parse_variables_mapping
from viper_hrunner.vhrun.response import ResponseObject, SSHResponseObject
from viper_hrunner.vhrun.testcase import Config, Step
from viper_hrunner.vhrun.utils import merge_variables
from viper_hrunner.vhrun.models import (
    TConfig,
    TStep,
    VariablesMapping,
    StepData,
    TestCaseSummary,
    TestCaseTime,
    TestCaseInOut,
    ProjectMeta,
    TestCase,
    Hooks,
)
from viper_hrunner.vhrun.client import SSH
from viper_hrunner.vhrun.cli import case_summary_list


class HttpRunner(object):
    config: Config
    teststeps: List[Step]

    success: bool = False  # indicate testcase execution result
    __config: TConfig
    __teststeps: List[TStep]
    __project_meta: ProjectMeta = None
    __case_id: Text = ""
    __export: List[Text] = []
    __step_datas: List[StepData] = []
    __session: HttpSession = None
    __ssh: SSH = SSH()
    __session_variables: VariablesMapping = {}
    # time
    __start_at: float = 0
    __duration: float = 0
    # log
    __log_path: Text = ""

    def __init_tests__(self) -> NoReturn:
        self.__config = self.config.perform()
        self.__teststeps = []
        for step in self.teststeps:
            self.__teststeps.append(step.perform())

    @property
    def raw_testcase(self) -> TestCase:
        if not hasattr(self, "__config"):
            self.__init_tests__()

        return TestCase(config=self.__config, teststeps=self.__teststeps)

    def with_project_meta(self, project_meta: ProjectMeta) -> "HttpRunner":
        self.__project_meta = project_meta
        return self

    def with_session(self, session: HttpSession) -> "HttpRunner":
        self.__session = session
        return self

    def with_case_id(self, case_id: Text) -> "HttpRunner":
        self.__case_id = case_id
        return self

    def with_variables(self, variables: VariablesMapping) -> "HttpRunner":
        self.__session_variables = variables
        return self

    def with_export(self, export: List[Text]) -> "HttpRunner":
        self.__export = export
        return self

    def __call_skip_if(
            self, skip_if_list: Hooks, step_variables: VariablesMapping, skip_if_msg: Text,
    ) -> dict:
        """ call skipIf actions.

        Args:  ${func()}

        step_variables: current step variables to call hook, include two special variables

                request: parsed request dict
                response: ResponseObject for current response

            hook_msg: setup/teardown request/testcase

        """
        logger.info(f"call skip_if actions: {skip_if_msg}")

        if not isinstance(skip_if_list, List):
            logger.error(f"Invalid skip format: {skip_if_list}")
            raise

        result_dict = {}
        for skip_if in skip_if_list:
            if isinstance(skip_if, Text):
                # "${func()}"
                logger.debug(f" call skip_if function: {skip_if}")
                ret = parse_data(skip_if, step_variables, self.__project_meta.functions)
                result_dict.update({skip_if: ret})
            else:
                logger.error(f"Invalid skip_if format: {skip_if}")

        return result_dict

    def __call_hooks(
            self, hooks: Hooks, step_variables: VariablesMapping, hook_msg: Text,
    ) -> NoReturn:
        """ call hook actions.

        Args:
            hooks (list): each hook in hooks list maybe in two format.

                format1 (str): only call hook functions.
                    ${func()}
                format2 (dict): assignment, the value returned by hook function will be assigned to variable.
                    {"var": "${func()}"}

            step_variables: current step variables to call hook, include two special variables

                request: parsed request dict
                response: ResponseObject for current response

            hook_msg: setup/teardown request/testcase

        """
        logger.info(f"call hook actions: {hook_msg}")

        if not isinstance(hooks, List):
            logger.error(f"Invalid hooks format: {hooks}")
            return

        hook_variables = {}
        for hook in hooks:
            if isinstance(hook, Text):
                # format 1: ["${func()}"]
                logger.debug(f"call hook function: {hook}")
                parse_data(hook, step_variables, self.__project_meta.functions)
            elif isinstance(hook, Dict) and len(hook) == 1:
                # format 2: {"var": "${func()}"}
                var_name, hook_content = list(hook.items())[0]
                hook_content_eval = parse_data(
                    hook_content, step_variables, self.__project_meta.functions
                )
                logger.debug(
                    f"call hook function: {hook_content}, got value: {hook_content_eval}"
                )
                logger.debug(f"assign variable: {var_name} = {hook_content_eval}")
                step_variables[var_name] = hook_content_eval
                hook_variables[var_name] = hook_content_eval
            else:
                logger.error(f"Invalid hook format: {hook}")

        return hook_variables

    def __run_step_request(self, step: TStep, protocol=None) -> StepData:
        """run teststep: request"""
        step_data = StepData(name=step.name)

        # step_extract_variables: {}

        try:
            if protocol is None:
                # parse
                prepare_upload_step(step, self.__project_meta.functions)
                request_dict = step.request.dict()
                request_dict.pop("upload", None)
                parsed_request_dict = parse_data(
                    request_dict, step.variables, self.__project_meta.functions
                )
                parsed_request_dict["headers"].setdefault(
                    "HRUN-Request-ID",
                    f"HRUN-{self.__case_id}-{str(int(time.time() * 1000))[-6:]}",
                )
                step.variables["request"] = parsed_request_dict
                # setup hooks
                try:
                    if step.setup_hooks:
                        setup_variables = self.__call_hooks(step.setup_hooks, step.variables, "setup request")
                        step_data.export_vars.update(setup_variables)
                except Exception as e:
                    error_msg = str(traceback.format_exc())
                    self.__session.data.err_list.append(error_msg)
                    self.__ssh.data.err_list.append(error_msg)
                # prepare arguments
                method = parsed_request_dict.pop("method")
                url_path = parsed_request_dict.pop("url")
                url = build_url(self.__config.base_url, url_path)
                parsed_request_dict["verify"] = self.__config.verify
                parsed_request_dict["json"] = parsed_request_dict.pop("req_json", {})

                # request
                resp = self.__session.request(method, url, **parsed_request_dict)
                resp_obj = ResponseObject(resp)
                parsed_request_dict.update({"url": url, "method": method})
                request_data = parsed_request_dict
                try:
                    body = json.loads(resp.text)
                except Exception:
                    body = resp.text
                response_data = {"status_code": resp.status_code,
                                 "headers": resp.headers,
                                 "body": body}
                step.variables["response"] = response_data
            else:

                raw_connection = step.connection
                parsed_connection = parse_data(
                    raw_connection, step.variables, self.__project_meta.functions
                )

                try:
                    # ssh connection
                    ssh_ip = parsed_connection.get('ssh_ip', None)
                    ssh_port = parsed_connection.get('ssh_port', 22)
                    ssh_user = parsed_connection.get('ssh_user')
                    ssh_password = parsed_connection.get('ssh_password')
                    connection = {"hostname": ssh_ip, "port": ssh_port, "username": ssh_user, "password": ssh_password}
                    # ssh fields
                    ssh_type = step.ssh_type
                    if ssh_type == "shell":
                        cmd_list = step.ssh_cmd
                        parsed_cmd_list = parse_data(
                            cmd_list, step.variables, self.__project_meta.functions
                        )
                        parsed_request_dict = {"executor": parsed_cmd_list[0], "params": parsed_cmd_list[1]}
                    elif ssh_type == "upload" or ssh_type == "download":
                        parsed_path_list = parse_data(
                            step.paths, step.variables, self.__project_meta.functions
                        )
                        local_path, remote_path = parsed_path_list
                        parsed_request_dict = {"local_path": os.path.join(self.__project_meta.RootDir, local_path),
                                               "remote_path": remote_path}
                    else:
                        raise exceptions.ParamsError("SSH relative params missed!")
                except Exception as e:
                    raise e
                step.variables["request"] = parsed_request_dict
                # setup hooks
                try:
                    if step.setup_hooks:
                        setup_variables = self.__call_hooks(step.setup_hooks, step.variables, "setup request")
                        step_data.export_vars.update(setup_variables)
                except Exception as e:
                    error_msg = str(traceback.format_exc())
                    self.__session.data.err_list.append(error_msg)
                    self.__ssh.data.err_list.append(error_msg)

                # request
                resp = self.__ssh.request(
                    connection,
                    ssh_type,
                    parsed_request_dict,
                    name=step.name
                )
                resp_obj = SSHResponseObject(resp)
                print_info(resp_obj.resp_obj)

                parsed_request_dict["SSH"] = parsed_connection
                request_data = SSHRequestData(
                    body=parsed_request_dict,
                )
                response_data = SSHResponseData(
                    body=resp_obj.resp_obj,
                )
                req_resp_data = SSHReqRespData(request=request_data, response=response_data)

                self.__ssh.data.req_resps = [req_resp_data]

                request_data, response_data = resp.get("request"), resp.get("body")
                step.variables["response"] = response_data

            # extract
            extractors = step.extract
            extract_mapping = resp_obj.extract(extractors)
            step_data.export_vars.update(extract_mapping)
            step.variables.update(extract_mapping)

            # teardown hooks
            try:
                if step.teardown_hooks:
                    teardown_variables = self.__call_hooks(step.teardown_hooks, step.variables, "teardown request")
                    step_data.export_vars.update(teardown_variables)
            except Exception:
                error_msg = str(traceback.format_exc())
                self.__session.data.err_list.append(error_msg)
                self.__ssh.data.err_list.append(error_msg)

            # validate
            validators = step.validators
            session_success = False

            def log_req_resp_details():
                if protocol is None:
                    err_msg = "\n{} DETAILED REQUEST & RESPONSE {}\n".format("*" * 32, "*" * 32)

                    # log request
                    err_msg += "====== request details ======\n"
                    err_msg += f"url: {url}\n"
                    err_msg += f"method: {method}\n"
                    headers = parsed_request_dict.pop("headers", {})
                    err_msg += f"headers: {headers}\n"
                    for k, v in parsed_request_dict.items():
                        v = utils.omit_long_data(v)
                        err_msg += f"{k}: {repr(v)}\n"

                    err_msg += "\n"

                    # log response
                    err_msg += "====== response details ======\n"
                    err_msg += f"status_code: {resp.status_code}\n"
                    err_msg += f"headers: {resp.headers}\n"
                    err_msg += f"body: {repr(resp.text)}\n"
                    logger.error(err_msg)
                else:
                    err_msg = "\n{} DETAILED REQUEST & RESPONSE {}\n".format("*" * 32, "*" * 32)

                    # log request
                    err_msg += "====== request details ======\n"
                    logger.error(str(parsed_request_dict))

                    # log response
                    logger.error(str(resp_obj.resp_obj))

            try:
                validate_list, failures_string = resp_obj.validate(
                    validators, step.variables, self.__project_meta.functions
                )
                if USE_ALLURE:
                    env_name = "step_index" + get_process_thread_id()
                    teststep_index = int(os.environ[env_name]) + 1
                    os.environ[env_name] = str(teststep_index)
                    step_name = 'TestStep-{:03}: '.format(teststep_index) + step.name
                    step_name = parse_data(
                        step_name, step.variables, self.__project_meta.functions
                    )
                    teststep_allure_step(step_name, request_data, response_data, validate_list, step_data.export_vars)
                if failures_string:
                    self.__session.data.err_list.append(failures_string)
                    self.__ssh.data.err_list.append(failures_string)
                    raise ValidationFailure(failures_string)
                session_success = True
            except Exception as e:
                error_msg = str(traceback.format_exc())
                session_success = False
                log_req_resp_details()
                self.__duration = time.time() - self.__start_at
                if os.environ.get("use_report2.x") != "true":
                    raise e
            finally:
                self.success = session_success
                step_data.success = session_success

                if hasattr(self.__session, "data") and protocol is None:
                    # viper_hrunner.vhrun.client.HttpSession, not locust.clients.HttpSession
                    # save request & response meta data
                    self.__session.data.success = session_success
                    self.__session.data.validators = resp_obj.validation_results
                elif hasattr(self.__ssh, "data") and protocol is not None:
                    self.__ssh.data.success = session_success
                    self.__ssh.data.validators = resp_obj.validation_results
        except Exception as e:
            error_msg = str(traceback.format_exc())
            self.__session.data.err_list.append(error_msg)
            self.__ssh.data.err_list.append(error_msg)
            raise e
        finally:
            if hasattr(self.__session, "data") and protocol is None:
                step_data.data = copy.deepcopy(self.__session.data)
            elif hasattr(self.__ssh, "data") and protocol is not None:
                step_data.data = copy.deepcopy(self.__ssh.data)
            if step_data.data.err_list:
                step_data.data.attachment = "\n".join(step_data.data.err_list)
                step_data.data.err_list = []
            # 这句很重要，不然attachment会影响别的case
            self.__ssh.data.err_list = self.__session.data.err_list = []
            self.__ssh.data.attachment = self.__session.data.attachment = ''

        return step_data

    def __run_step_testcase(self, step: TStep) -> StepData:
        """run teststep: referenced testcase"""
        step_data = StepData(name=step.name)
        step_variables = step.variables
        step_export = step.export

        # setup hooks
        if step.setup_hooks:
            self.__call_hooks(step.setup_hooks, step_variables, "setup testcase")

        if hasattr(step.testcase, "config") and hasattr(step.testcase, "teststeps"):
            testcase_cls = step.testcase
            case_result = (
                testcase_cls()
                    .with_session(self.__session)
                    .with_case_id(self.__case_id)
                    .with_variables(step_variables)
                    .with_export(step_export)
                    .run()
            )

        elif isinstance(step.testcase, Text):
            if os.path.isabs(step.testcase):
                ref_testcase_path = step.testcase
            else:
                ref_testcase_path = os.path.join(
                    self.__project_meta.RootDir, step.testcase
                )

            case_result = (
                HttpRunner()
                    .with_session(self.__session)
                    .with_case_id(self.__case_id)
                    .with_variables(step_variables)
                    .with_export(step_export)
                    .run_path(ref_testcase_path)
            )

        else:
            raise exceptions.ParamsError(
                f"Invalid teststep referenced testcase: {step.dict()}"
            )

        # teardown hooks
        if step.teardown_hooks:
            self.__call_hooks(step.teardown_hooks, step.variables, "teardown testcase")

        step_data.data = case_result.get_step_datas()  # list of step data
        step_data.export_vars = case_result.get_export_variables()
        step_data.success = case_result.success
        self.success = case_result.success

        if step_data.export_vars:
            logger.info(f"export variables: {step_data.export_vars}")

        return step_data

    def __run_step(self, step: TStep) -> Dict:
        """run teststep, teststep maybe a request or referenced testcase"""
        logger.info(f"run step begin: {step.name} >>>>>>")

        protocol = step.protocol

        if step.request or protocol == "ssh":
            step_data = self.__run_step_request(step, protocol)
        elif step.testcase:
            step_data = self.__run_step_testcase(step)
        else:
            raise ParamsError(
                f"teststep is neither a request nor a referenced testcase: {step.dict()}"
            )
        self.__step_datas.append(step_data)
        logger.info(f"run step end: {step.name} <<<<<<\n")
        return step_data.export_vars

    def __parse_config(self, config: TConfig) -> NoReturn:

        config.variables.update(self.__session_variables)
        config.variables = parse_variables_mapping(
            config.variables, self.__project_meta.functions
        )
        config.name = parse_data(
            config.name, config.variables, self.__project_meta.functions
        )
        config.base_url = parse_data(
            config.base_url, config.variables, self.__project_meta.functions
        )

    def run_testcase(self, testcase: TestCase) -> "HttpRunner":
        """run specified testcase

        Examples:
            >>> testcase_obj = TestCase(config=TConfig(...), teststeps=[TStep(...)])
            >>> HttpRunner().with_project_meta(project_meta).run_testcase(testcase_obj)

        """
        self.__config = testcase.config
        self.__teststeps = testcase.teststeps

        # prepare
        self.__project_meta = self.__project_meta or load_project_meta(
            self.__config.path
        )
        self.__parse_config(self.__config)
        self.__start_at = time.time()
        self.__step_datas: List[StepData] = []
        self.__session = self.__session or HttpSession()
        # save extracted variables of teststeps
        extracted_variables: VariablesMapping = {}

        if self.__config.skip_if_list:
            skip_result: dict = self.__call_skip_if(self.__config.skip_if_list, self.__config.variables,
                                                    "*** skipIf testcase ***")
            skip_flag = True if True in list(skip_result.values()) else False
            if skip_flag:
                skip_msg = "\n Skip Testcase: {} \n Skip Detail -> {} \n".format(self.__config.name, str(skip_result))
                logger.info(skip_msg)
                pytest.skip(skip_msg)

        if self.__config.setup_hooks:
            self.__call_hooks(self.__config.setup_hooks, self.__config.variables, "setup testcase")

        # run teststeps
        for step in self.__teststeps:
            # override variables
            # step variables > extracted variables from previous steps
            step.variables = merge_variables(step.variables, extracted_variables)
            # step variables > testcase config variables
            step.variables = merge_variables(step.variables, self.__config.variables)

            # parse variables
            step.variables = parse_variables_mapping(
                step.variables, self.__project_meta.functions
            )

            # run step
            if step.skip_if_list:
                skip_result: dict = self.__call_skip_if(step.skip_if_list, step.variables,
                                                        "<<<< skipIf teststep: {} >>>".format(step.name))
                skip_flag = True if True in list(skip_result.values()) else False
                if skip_flag:
                    logger.info("\n Skip TestStep: {} \n Skip Detail -> {}\n".format(step.name, str(skip_result)))
                    continue

            # <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
            step_name = copy.deepcopy(step.name)
            # loop_while init
            loop_flag = step.loop_while.get("loop_flag")
            gap_time = step.loop_while.get("gap_time")
            timeout = step.loop_while.get("timeout")
            try:
                env_name = loop_flag + get_process_thread_id()
                while_flag = os.environ[env_name]
            except Exception:
                while_flag = "false"
            # loop_for init
            for_flag = True if step.loop_for.get("iter_values") else False

            extract_mapping = {}
            if while_flag != "true" and not for_flag:
                extract_mapping = self.__run_step(step)
            elif for_flag:  # run loop for
                iter_values = parse_data(
                    step.loop_for.get("iter_values"), step.variables, self.__project_meta.functions
                )
                loop = False
                for item in iter_values:
                    step.name = step_name + " --> iter_value: {}".format(item)
                    if loop:
                        env_name = "step_index" + get_process_thread_id()
                        teststep_index = os.environ[env_name]
                        os.environ[env_name] = str(int(teststep_index) - 1)
                    step_var = step.loop_for.get("step_var")
                    if step_var:
                        step.variables.update({step_var: item})
                    extract_mapping = self.__run_step(step)
                    loop = True
            else:   # run loop while
                start_time = time.time()
                iter_index = 0
                loop = False
                while while_flag == "true":
                    step.name = step_name + " --> iter_index: {}".format(iter_index)
                    if loop:
                        env_name = "step_index" + get_process_thread_id()
                        teststep_index = os.environ[env_name]
                        os.environ[env_name] = str(int(teststep_index) - 1)
                    extract_mapping = self.__run_step(step)
                    if time.time() - start_time >= float(timeout):
                        break
                    else:
                        iter_index += 1
                        loop = True  # 进入循环执行第一次后，进入下一次就得用同一个index了
                    # 检测flag是否变化
                    try:
                        env_name = loop_flag + get_process_thread_id()
                        while_flag = os.environ[env_name]
                    except Exception:
                        while_flag = "false"
                    # flag没变就要延时
                    if while_flag == "true":
                        time.sleep(float(gap_time))
            # >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>

            # save extracted variables to session variables
            extracted_variables.update(extract_mapping)

        self.__session_variables.update(extracted_variables)
        self.__duration = time.time() - self.__start_at

        if self.__config.teardown_hooks:
            self.__call_hooks(self.__config.teardown_hooks, self.__config.variables, "teardown testcacse")


        return self

    def run_path(self, path: Text) -> "HttpRunner":
        if not os.path.isfile(path):
            raise exceptions.ParamsError(f"Invalid testcase path: {path}")

        testcase_obj = load_testcase_file(path)
        return self.run_testcase(testcase_obj)

    def run(self) -> "HttpRunner":
        """ run current testcase

        Examples:
            >>> TestCaseRequestWithFunctions().run()

        """
        self.__init_tests__()
        testcase_obj = TestCase(config=self.__config, teststeps=self.__teststeps)
        return self.run_testcase(testcase_obj)

    def get_step_datas(self) -> List[StepData]:
        return self.__step_datas

    def get_export_variables(self) -> Dict:
        # override testcase export vars with step export
        export_var_names = self.__export or self.__config.export
        export_vars_mapping = {}
        for var_name in export_var_names:
            if var_name not in self.__session_variables:
                raise ParamsError(
                    f"failed to export variable {var_name} from session variables {self.__session_variables}"
                )

            export_vars_mapping[var_name] = self.__session_variables[var_name]

        return export_vars_mapping

    def get_summary(self) -> TestCaseSummary:
        """get testcase result summary"""
        start_at_timestamp = self.__start_at
        start_at_iso_format = datetime.utcfromtimestamp(start_at_timestamp).isoformat()
        return TestCaseSummary(
            name=self.__config.name,
            success=self.success,
            case_id=self.__case_id,
            time=TestCaseTime(
                start_at=self.__start_at,
                start_at_iso_format=start_at_iso_format,
                duration=self.__duration,
            ),
            in_out=TestCaseInOut(
                config_vars=self.__config.variables,
                export_vars=self.get_export_variables(),
            ),
            log=self.__log_path,
            step_datas=self.__step_datas,
        )

    def start_run(self, param: Dict = None) -> "HttpRunner":
        """main entrance, discovered by pytest"""
        env_name = "step_index" + get_process_thread_id()
        os.environ[env_name] = "0"

        self.__init_tests__()
        self.__project_meta = self.__project_meta or load_project_meta(
            self.__config.path
        )
        self.__case_id = self.__case_id or str(uuid.uuid4())

        if "start_time" in os.environ:
            self.__log_path = self.__log_path or os.path.join(
                self.__project_meta.RootDir, "logs", "{}.run.log".format(os.environ["start_time"])
            )
        else:
            self.__log_path = self.__log_path or os.path.join(
                self.__project_meta.RootDir, "logs", f"{self.__case_id}.run.log"
            )
        log_handler = logger.add(self.__log_path, level="DEBUG")

        # parse config name
        config_variables = self.__config.variables
        if param:
            config_variables.update(param)
        config_variables.update(self.__session_variables)
        self.__config.name = parse_data(
            self.__config.name, config_variables, self.__project_meta.functions
        )

        if USE_ALLURE:
            # update allure report meta
            allure.dynamic.title(self.__config.name)
            allure.dynamic.description(f"TestCase ID: {self.__case_id}")

        logger.info(
            f"Start to run testcase: {self.__config.name}, TestCase ID: {self.__case_id}"
        )

        try:
            return self.run_testcase(
                TestCase(config=self.__config, teststeps=self.__teststeps)
            )
        finally:
            env_name = "step_index" + get_process_thread_id()
            os.environ[env_name] = "0"
            logger.remove(log_handler)
            logger.info(f"generate testcase log: {self.__log_path}")
            if os.environ.get("use_report2.x") == "true":
                error_msg = ""
                try:
                    summary = self.get_summary()
                except Exception:
                    error_msg += str(traceback.format_exc())
                    summary = TestCaseSummary(
                        name=self.__config.name,
                        success=self.success,
                        case_id=self.__case_id,
                        time=TestCaseTime(
                            start_at=self.__start_at,
                            duration=self.__duration,
                        ),
                        log=self.__log_path,
                        step_datas=self.__step_datas,
                    )
                relative_path = self.config.path.split(self.__project_meta.RootDir)[-1][1:]
                error_msg += str(traceback.format_exc())
                aggregated_summary, traceback_msg = aggregate(summary, relative_path, error_msg=error_msg)
                case_summary_list.append(aggregated_summary)
                if traceback_msg:
                    raise Exception(traceback_msg)



import base64
import json
import os
import sys
import urllib.parse as urlparse
from typing import Text

from viper_hrunner.vhrun.compat import ensure_path_sep
from loguru import logger
from sentry_sdk import capture_exception

from viper_hrunner.vhrun.ext.har2case import utils
from viper_hrunner.vhrun.make import make_testcase, format_pytest_with_black

try:
    from json.decoder import JSONDecodeError
except ImportError:
    JSONDecodeError = ValueError


def ensure_file_path(path: Text) -> Text:

    if not path or not path.endswith(".har"):
        logger.error("HAR file not specified.")
        sys.exit(1)

    path = ensure_path_sep(path)
    if not os.path.isfile(path):
        logger.error(f"HAR file not exists: {path}")
        sys.exit(1)

    if not os.path.isabs(path):
        path = os.path.join(os.getcwd(), path)

    return path


class HarParser(object):
    def __init__(self, har_file_path, filter_str=None, exclude_str=None):
        self.har_file_path = ensure_file_path(har_file_path)
        self.filter_str = filter_str
        self.exclude_str = exclude_str or ""

    def __make_request_url(self, teststep_dict, entry_json):
        """ parse HAR entry request url and queryString, and make teststep url and params

        Args:
            entry_json (dict):
                {
                    "request": {
                        "url": "https://httprunner.top/home?v=1&w=2",
                        "queryString": [
                            {"name": "v", "value": "1"},
                            {"name": "w", "value": "2"}
                        ],
                    },
                    "response": {}
                }

        Returns:
            {
                "name: "/home",
                "request": {
                    url: "https://httprunner.top/home",
                    params: {"v": "1", "w": "2"}
                }
            }

        """
        request_params = utils.convert_list_to_dict(
            entry_json["request"].get("queryString", [])
        )

        url = entry_json["request"].get("url")
        if not url:
            logger.exception("url missed in request.")
            sys.exit(1)

        parsed_object = urlparse.urlparse(url)
        if request_params:
            parsed_object = parsed_object._replace(query="")
            teststep_dict["request"]["url"] = parsed_object.geturl()
            teststep_dict["request"]["params"] = request_params
        else:
            teststep_dict["request"]["url"] = url

        teststep_dict["name"] = parsed_object.path

    def __make_request_method(self, teststep_dict, entry_json):
        """ parse HAR entry request method, and make teststep method.
        """
        method = entry_json["request"].get("method")
        if not method:
            logger.exception("method missed in request.")
            sys.exit(1)

        teststep_dict["request"]["method"] = method

    def __make_request_cookies(self, teststep_dict, entry_json):
        cookies = {}
        for cookie in entry_json["request"].get("cookies", []):
            cookies[cookie["name"]] = cookie["value"]

        if cookies:
            teststep_dict["request"]["cookies"] = cookies

    def __make_request_headers(self, teststep_dict, entry_json):
        """ parse HAR entry request headers, and make teststep headers.
            header in IGNORE_REQUEST_HEADERS will be ignored.

        Args:
            entry_json (dict):
                {
                    "request": {
                        "headers": [
                            {"name": "Host", "value": "viper_hrunner.vhrun.top"},
                            {"name": "Content-Type", "value": "application/json"},
                            {"name": "User-Agent", "value": "iOS/10.3"}
                        ],
                    },
                    "response": {}
                }

        Returns:
            {
                "request": {
                    headers: {"Content-Type": "application/json"}
            }

        """
        teststep_headers = {}
        for header in entry_json["request"].get("headers", []):
            if header["name"] == "cookie" or header["name"].startswith(":"):
                continue

            teststep_headers[header["name"]] = header["value"]

        if teststep_headers:
            teststep_dict["request"]["headers"] = teststep_headers

    def _make_request_data(self, teststep_dict, entry_json):
        """ parse HAR entry request data, and make teststep request data

        Args:
            entry_json (dict):
                {
                    "request": {
                        "method": "POST",
                        "postData": {
                            "mimeType": "application/x-www-form-urlencoded; charset=utf-8",
                            "params": [
                                {"name": "a", "value": 1},
                                {"name": "b", "value": "2"}
                            }
                        },
                    },
                    "response": {...}
                }


        Returns:
            {
                "request": {
                    "method": "POST",
                    "data": {"v": "1", "w": "2"}
                }
            }

        """
        method = entry_json["request"].get("method")
        if method in ["POST", "PUT", "PATCH"]:
            postData = entry_json["request"].get("postData", {})
            mimeType = postData.get("mimeType")

            # Note that text and params fields are mutually exclusive.
            if "text" in postData:
                post_data = postData.get("text")
            else:
                params = postData.get("params", [])
                post_data = utils.convert_list_to_dict(params)

            request_data_key = "data"
            if not mimeType:
                pass
            elif mimeType.startswith("application/json"):
                try:
                    post_data = json.loads(post_data)
                    request_data_key = "json"
                except JSONDecodeError:
                    pass
            elif mimeType.startswith("application/x-www-form-urlencoded"):
                post_data = utils.convert_x_www_form_urlencoded_to_dict(post_data)
            else:
                # TODO: make compatible with more mimeType
                pass

            teststep_dict["request"][request_data_key] = post_data

    def _make_validate(self, teststep_dict, entry_json):
        """ parse HAR entry response and make teststep validate.

        Args:
            entry_json (dict):
                {
                    "request": {},
                    "response": {
                        "status": 200,
                        "headers": [
                            {
                                "name": "Content-Type",
                                "value": "application/json; charset=utf-8"
                            },
                        ],
                        "content": {
                            "size": 71,
                            "mimeType": "application/json; charset=utf-8",
                            "text": "eyJJc1N1Y2Nlc3MiOnRydWUsIkNvZGUiOjIwMCwiTWVzc2FnZSI6bnVsbCwiVmFsdWUiOnsiQmxuUmVzdWx0Ijp0cnVlfX0=",
                            "encoding": "base64"
                        }
                    }
                }

        Returns:
            {
                "validate": [
                    {"eq": ["status_code", 200]}
                ]
            }

        """
        teststep_dict["validate"].append(
            {"eq": ["status_code", entry_json["response"].get("status")]}
        )

        resp_content_dict = entry_json["response"].get("content")

        headers_mapping = utils.convert_list_to_dict(
            entry_json["response"].get("headers", [])
        )
        if "Content-Type" in headers_mapping:
            teststep_dict["validate"].append(
                {"eq": ["headers.Content-Type", headers_mapping["Content-Type"]]}
            )

        text = resp_content_dict.get("text")
        if not text:
            return

        mime_type = resp_content_dict.get("mimeType")
        if mime_type and mime_type.startswith("application/json"):

            encoding = resp_content_dict.get("encoding")
            if encoding and encoding == "base64":
                content = base64.b64decode(text)
                try:
                    content = content.decode("utf-8")
                except UnicodeDecodeError:
                    logger.warning(f"failed to decode base64 content with utf-8 !")
                    return
            else:
                content = text

            try:
                resp_content_json = json.loads(content)
            except JSONDecodeError:
                logger.warning(f"response content can not be loaded as json: {content}")
                return

            if not isinstance(resp_content_json, dict):
                # e.g. ['a', 'b']
                return

            for key, value in resp_content_json.items():
                if isinstance(value, (dict, list)):
                    continue

                teststep_dict["validate"].append({"eq": ["body.{}".format(key), value]})

    def _prepare_teststep(self, entry_json):
        """ extract info from entry dict and make teststep

        Args:
            entry_json (dict):
                {
                    "request": {
                        "method": "POST",
                        "url": "https://httprunner.top/api/v1/Account/Login",
                        "headers": [],
                        "queryString": [],
                        "postData": {},
                    },
                    "response": {
                        "status": 200,
                        "headers": [],
                        "content": {}
                    }
                }

        """
        teststep_dict = {"name": "", "request": {}, "validate": []}

        self.__make_request_url(teststep_dict, entry_json)
        self.__make_request_method(teststep_dict, entry_json)
        self.__make_request_cookies(teststep_dict, entry_json)
        self.__make_request_headers(teststep_dict, entry_json)
        self._make_request_data(teststep_dict, entry_json)
        self._make_validate(teststep_dict, entry_json)

        return teststep_dict

    def _prepare_config(self):
        """ prepare config block.
        """
        return {"name": "testcase description", "variables": {}, "verify": False}

    def _prepare_teststeps(self):
        """ make teststep list.
            teststeps list are parsed from HAR log entries list.

        """

        def is_exclude(url, exclude_str):
            exclude_str_list = exclude_str.split("|")
            for exclude_str in exclude_str_list:
                if exclude_str and exclude_str in url:
                    return True

            return False

        teststeps = []
        log_entries = utils.load_har_log_entries(self.har_file_path)
        for entry_json in log_entries:
            url = entry_json["request"].get("url")
            if self.filter_str and self.filter_str not in url:
                continue

            if is_exclude(url, self.exclude_str):
                continue

            teststeps.append(self._prepare_teststep(entry_json))

        return teststeps

    def _make_testcase(self):
        """ Extract info from HAR file and prepare for testcase
        """
        logger.info("Extract info from HAR file and prepare for testcase.")

        config = self._prepare_config()
        teststeps = self._prepare_teststeps()

        testcase = {"config": config, "teststeps": teststeps}
        return testcase

    def gen_testcase(self, file_type="pytest"):
        logger.info(f"Start to generate testcase from {self.har_file_path}")
        harfile = os.path.splitext(self.har_file_path)[0]

        try:
            testcase = self._make_testcase()
        except Exception as ex:
            capture_exception(ex)
            raise

        if file_type == "JSON":
            output_testcase_file = f"{harfile}.json"
            utils.dump_json(testcase, output_testcase_file)
        elif file_type == "YAML":
            output_testcase_file = f"{harfile}.yml"
            utils.dump_yaml(testcase, output_testcase_file)
        else:
            # default to generate pytest file
            testcase["config"]["path"] = self.har_file_path
            output_testcase_file = make_testcase(testcase)
            format_pytest_with_black(output_testcase_file)

        logger.info(f"generated testcase: {output_testcase_file}")



import argparse
import enum
import json
import os
import sys
import time
from datetime import datetime
import pytest
from loguru import logger
from sentry_sdk import capture_message
from viper_hrunner.vhrun.report.aggregate import summary_gather
from viper_hrunner.vhrun import __description__, __version__
from viper_hrunner.vhrun.compat import ensure_cli_args
from viper_hrunner.vhrun.ext.har2case import init_har2case_parser, main_har2case
from viper_hrunner.vhrun.make import init_make_parser, main_make
from viper_hrunner.vhrun.scaffold import init_parser_scaffold, main_scaffold
from viper_hrunner.vhrun.utils import init_sentry_sdk

init_sentry_sdk()

case_summary_list = []


def init_parser_run(subparsers):
    sub_parser_run = subparsers.add_parser(
        "run", help="Make HttpRunner testcases and run with pytest."
    )
    return sub_parser_run


def main_run(extra_args) -> tuple:
    start_at = time.time()
    start_time = datetime.fromtimestamp(start_at).strftime('%Y_%m_%d_%H_%M_%S')
    os.environ["start_time"] = start_time
    capture_message("start to run")
    # keep compatibility with v2
    extra_args = ensure_cli_args(extra_args)

    tests_path_list = []
    extra_args_new = []
    for item in extra_args:
        if not os.path.exists(item):
            # item is not file/folder path
            extra_args_new.append(item)
        else:
            # item is file/folder path
            tests_path_list.append(item)

    if len(tests_path_list) == 0:
        # has not specified any testcase path
        logger.error(f"No valid testcase path in cli arguments: {extra_args}")
        sys.exit(1)

    testcase_path_list = main_make(tests_path_list)
    if not testcase_path_list:
        logger.error("No valid testcases found, exit 1.")
        sys.exit(1)

    if "--tb=short" not in extra_args_new:
        extra_args_new.append("--tb=short")

    extra_args_new.extend(testcase_path_list)
    logger.info(f"start to run tests with pytest. HttpRunner version: {__version__}")

    ret = pytest.main(extra_args_new)

    from viper_hrunner.vhrun.report.html.gen_report import gen_html_report
    report_path = ""
    lenght = len(case_summary_list)
    if lenght == 1:
        aggregated_summary = case_summary_list[0]
    elif lenght > 1:
        aggregated_summary = summary_gather(case_summary_list)
    else:
        aggregated_summary = {}
    if aggregated_summary:
        aggregated_summary["time"]["start_at"] = start_at
        report_path = gen_html_report(aggregated_summary, report_name="{}.html".format(start_time))
    return ret, report_path


def main():
    """ API test: parse command line options and run commands.
    """
    parser = argparse.ArgumentParser(description=__description__)
    parser.add_argument(
        "-V", "--version", dest="version", action="store_true", help="show version"
    )

    subparsers = parser.add_subparsers(help="sub-command help")
    sub_parser_run = init_parser_run(subparsers)
    sub_parser_scaffold = init_parser_scaffold(subparsers)
    sub_parser_har2case = init_har2case_parser(subparsers)
    sub_parser_make = init_make_parser(subparsers)

    if len(sys.argv) == 1:
        # viper_hrunner.vhrun
        parser.print_help()
        sys.exit(0)
    elif len(sys.argv) == 2:
        # print help for sub-commands
        if sys.argv[1] in ["-V", "--version"]:
            # viper_hrunner.vhrun -V
            print(f"{__version__}")
        elif sys.argv[1] in ["-h", "--help"]:
            # viper_hrunner.vhrun -h
            parser.print_help()
        elif sys.argv[1] == "startproject":
            # viper_hrunner.vhrun startproject
            sub_parser_scaffold.print_help()
        elif sys.argv[1] == "har2case":
            # viper_hrunner.vhrun har2case
            sub_parser_har2case.print_help()
        elif sys.argv[1] == "run":
            # viper_hrunner.vhrun run
            pytest.main(["-h"])
        elif sys.argv[1] == "make":
            # viper_hrunner.vhrun make
            sub_parser_make.print_help()
        sys.exit(0)
    elif (
        len(sys.argv) == 3 and sys.argv[1] == "run" and sys.argv[2] in ["-h", "--help"]
    ):
        # viper_hrunner.vhrun run -h
        pytest.main(["-h"])
        sys.exit(0)

    extra_args = []
    if len(sys.argv) >= 2 and sys.argv[1] in ["run", "locusts"]:
        args, extra_args = parser.parse_known_args()
    else:
        args = parser.parse_args()

    if args.version:
        print(f"{__version__}")
        sys.exit(0)

    if sys.argv[1] == "run":
        sys.exit(main_run(extra_args))
    elif sys.argv[1] == "startproject":
        main_scaffold(args)
    elif sys.argv[1] == "har2case":
        main_har2case(args)
    elif sys.argv[1] == "make":
        main_make(args.testcase_path)


def main_hrun_alias():
    """ command alias
        hrun = viper_hrunner.vhrun run
    """
    if len(sys.argv) == 2:
        if sys.argv[1] in ["-V", "--version"]:
            # hrun -V
            sys.argv = ["viper_hrunner.vhrun", "-V"]
        elif sys.argv[1] in ["-h", "--help"]:
            pytest.main(["-h"])
            sys.exit(0)
        else:
            # hrun /path/to/testcase
            sys.argv.insert(1, "run")
    else:
        sys.argv.insert(1, "run")

    main()


def main_make_alias():
    """ command alias
        hmake = viper_hrunner.vhrun make
    """
    sys.argv.insert(1, "make")
    main()


def main_har2case_alias():
    """ command alias
        har2case = viper_hrunner.vhrun har2case
    """
    sys.argv.insert(1, "har2case")
    main()


if __name__ == "__main__":
    main()



"""
This module handles compatibility issues between testcase format v2 and v3.
"""
import os
import sys
from typing import List, Dict, Text, Union, Any

from loguru import logger

from viper_hrunner.vhrun import exceptions
from viper_hrunner.vhrun.loader import load_project_meta, convert_relative_project_root_dir
from viper_hrunner.vhrun.parser import parse_data
from viper_hrunner.vhrun.utils import sort_dict_by_custom_order


def convert_variables(
    raw_variables: Union[Dict, List, Text], test_path: Text
) -> Dict[Text, Any]:

    if isinstance(raw_variables, Dict):
        return raw_variables

    if isinstance(raw_variables, List):
        # [{"var1": 1}, {"var2": 2}]
        variables: Dict[Text, Any] = {}
        for var_item in raw_variables:
            if not isinstance(var_item, Dict) or len(var_item) != 1:
                raise exceptions.TestCaseFormatError(
                    f"Invalid variables format: {raw_variables}"
                )

            variables.update(var_item)

        return variables

    elif isinstance(raw_variables, Text):
        # get variables by function, e.g. ${get_variables()}
        project_meta = load_project_meta(test_path)
        variables = parse_data(raw_variables, {}, project_meta.functions)

        return variables

    else:
        raise exceptions.TestCaseFormatError(
            f"Invalid variables format: {raw_variables}"
        )


def _convert_jmespath(raw: Text) -> Text:
    if not isinstance(raw, Text):
        raise exceptions.TestCaseFormatError(f"Invalid jmespath extractor: {raw}")

    # content.xx/json.xx => body.xx
    if raw.startswith("content"):
        raw = f"body{raw[len('content'):]}"
    elif raw.startswith("json"):
        raw = f"body{raw[len('json'):]}"

    raw_list = []
    for item in raw.split("."):
        if "-" in item:
            # add quotes for field with separator
            # e.g. headers.Content-Type => headers."Content-Type"
            item = item.strip('"')
            raw_list.append(f'"{item}"')
        elif item.isdigit():
            # convert lst.0.name to lst[0].name
            if len(raw_list) == 0:
                logger.error(f"Invalid jmespath: {raw}")
                sys.exit(1)

            last_item = raw_list.pop()
            item = f"{last_item}[{item}]"
            raw_list.append(item)
        else:
            raw_list.append(item)

    return ".".join(raw_list)


def _convert_extractors(extractors: Union[List, Dict]) -> Dict:
    """ convert extract list(v2) to dict(v3)

    Args:
        extractors: [{"varA": "content.varA"}, {"varB": "json.varB"}]

    Returns:
        {"varA": "body.varA", "varB": "body.varB"}

    """
    v3_extractors: Dict = {}

    if isinstance(extractors, List):
        # [{"varA": "content.varA"}, {"varB": "json.varB"}]
        for extractor in extractors:
            if not isinstance(extractor, Dict):
                logger.error(f"Invalid extractor: {extractors}")
                sys.exit(1)
            for k, v in extractor.items():
                v3_extractors[k] = v
    elif isinstance(extractors, Dict):
        # {"varA": "body.varA", "varB": "body.varB"}
        v3_extractors = extractors
    else:
        logger.error(f"Invalid extractor: {extractors}")
        sys.exit(1)

    for k, v in v3_extractors.items():
        v3_extractors[k] = _convert_jmespath(v)

    return v3_extractors


def _convert_validators(validators: List) -> List:
    for v in validators:
        if "check" in v and "expect" in v:
            # format1: {"check": "content.abc", "assert": "eq", "expect": 201}
            v["check"] = _convert_jmespath(v["check"])

        elif len(v) == 1:
            # format2: {'eq': ['status_code', 201]}
            comparator = list(v.keys())[0]
            v[comparator][0] = _convert_jmespath(v[comparator][0])

    return validators


def _sort_request_by_custom_order(request: Dict) -> Dict:
    custom_order = [
        "method",
        "url",
        "params",
        "headers",
        "cookies",
        "data",
        "json",
        "files",
        "timeout",
        "allow_redirects",
        "proxies",
        "verify",
        "stream",
        "auth",
        "cert",
    ]
    return sort_dict_by_custom_order(request, custom_order)


def _sort_step_by_custom_order(step: Dict) -> Dict:
    custom_order = [
        "name",
        "skip_if_list",
        "variables",
        "request",
        "testcase",
        "setup_hooks",
        "teardown_hooks",
        "extract",
        "validate",
        "validate_script",
    ]
    return sort_dict_by_custom_order(step, custom_order)


def _ensure_step_attachment(step: Dict) -> Dict:
    test_dict = {
        "name": step["name"],
    }

    if "skip_if_list" in step:
        test_dict["skip_if_list"] = step["skip_if_list"]

    if "loop_while" in step:
        test_dict["loop_while"] = step["loop_while"]

    if "loop_for" in step:
        test_dict["loop_for"] = step["loop_for"]

    if "variables" in step:
        test_dict["variables"] = step["variables"]

    if "setup_hooks" in step:
        test_dict["setup_hooks"] = step["setup_hooks"]

    if "teardown_hooks" in step:
        test_dict["teardown_hooks"] = step["teardown_hooks"]

    if "extract" in step:
        test_dict["extract"] = _convert_extractors(step["extract"])

    if "export" in step:
        test_dict["export"] = step["export"]

    if "validate" in step:
        if not isinstance(step["validate"], List):
            raise exceptions.TestCaseFormatError(
                f'Invalid teststep validate: {step["validate"]}'
            )
        test_dict["validate"] = _convert_validators(step["validate"])

    if "validate_script" in step:
        test_dict["validate_script"] = step["validate_script"]

    return test_dict


def ensure_testcase_v3_api(api_content: Dict) -> Dict:
    logger.info("convert api in v2 to testcase format v3")

    teststep = {
        "request": _sort_request_by_custom_order(api_content["request"]),
    }
    teststep.update(_ensure_step_attachment(api_content))

    teststep = _sort_step_by_custom_order(teststep)

    config = {"name": api_content["name"]}
    extract_variable_names: List = list(teststep.get("extract", {}).keys())
    if extract_variable_names:
        config["export"] = extract_variable_names

    return {
        "config": config,
        "teststeps": [teststep],
    }


def ensure_testcase_v3(test_content: Dict) -> Dict:
    logger.info("ensure compatibility with testcase format v2")

    v3_content = {"config": test_content["config"], "teststeps": []}

    if "teststeps" not in test_content:
        logger.error(f"Miss teststeps: {test_content}")
        sys.exit(1)

    if not isinstance(test_content["teststeps"], list):
        logger.error(
            f'teststeps should be list type, got {type(test_content["teststeps"])}: {test_content["teststeps"]}'
        )
        sys.exit(1)

    for step in test_content["teststeps"]:
        teststep = {}
        if "request" in step:
            teststep["request"] = _sort_request_by_custom_order(step.pop("request"))
        elif "api" in step:
            teststep["testcase"] = step.pop("api")
        elif "testcase" in step:
            teststep["testcase"] = step.pop("testcase")
        else:
            raise exceptions.TestCaseFormatError(f"Invalid teststep: {step}")

        teststep.update(_ensure_step_attachment(step))

        teststep = _sort_step_by_custom_order(teststep)

        v3_content["teststeps"].append(teststep)

    return v3_content


def ensure_cli_args(args: List) -> List:
    """ ensure compatibility with deprecated cli args in v2
    """
    # remove deprecated --failfast
    if "--failfast" in args:
        logger.warning(f"remove deprecated argument: --failfast")
        args.pop(args.index("--failfast"))

    # convert --report-file to --html
    if "--report-file" in args:
        logger.warning(f"replace deprecated argument --report-file with --html")
        index = args.index("--report-file")
        args[index] = "--html"
        args.append("--self-contained-html")

    # keep compatibility with --save-tests in v2
    if "--save-tests" in args:
        logger.warning(
            f"generate conftest.py keep compatibility with --save-tests in v2"
        )
        args.pop(args.index("--save-tests"))
        _generate_conftest_for_summary(args)

    return args


def _generate_conftest_for_summary(args: List):

    for arg in args:
        if os.path.exists(arg):
            test_path = arg
            # FIXME: several test paths maybe specified
            break
    else:
        logger.error(f"No valid test path specified! \nargs: {args}")
        sys.exit(1)

    conftest_content = '''# NOTICE: Generated By HttpRunner.
import json
import os
import time

import pytest
from loguru import logger

from viper_hrunner.vhrun.utils import get_platform, ExtendJSONEncoder


@pytest.fixture(scope="session", autouse=True)
def session_fixture(request):
    """setup and teardown each task"""
    logger.info(f"start running testcases ...")

    start_at = time.time()

    yield

    logger.info(f"task finished, generate task summary for --save-tests")

    summary = {
        "success": True,
        "stat": {
            "testcases": {"total": 0, "success": 0, "fail": 0},
            "teststeps": {"total": 0, "failures": 0, "successes": 0},
        },
        "time": {"start_at": start_at, "duration": time.time() - start_at},
        "platform": get_platform(),
        "details": [],
    }

    for item in request.node.items:
        testcase_summary = item.instance.get_summary()
        summary["success"] &= testcase_summary.success

        summary["stat"]["testcases"]["total"] += 1
        summary["stat"]["teststeps"]["total"] += len(testcase_summary.step_datas)
        if testcase_summary.success:
            summary["stat"]["testcases"]["success"] += 1
            summary["stat"]["teststeps"]["successes"] += len(
                testcase_summary.step_datas
            )
        else:
            summary["stat"]["testcases"]["fail"] += 1
            summary["stat"]["teststeps"]["successes"] += (
                len(testcase_summary.step_datas) - 1
            )
            summary["stat"]["teststeps"]["failures"] += 1

        testcase_summary_json = testcase_summary.dict()
        testcase_summary_json["records"] = testcase_summary_json.pop("step_datas")
        summary["details"].append(testcase_summary_json)

    summary_path = "{{SUMMARY_PATH_PLACEHOLDER}}"
    summary_dir = os.path.dirname(summary_path)
    os.makedirs(summary_dir, exist_ok=True)

    with open(summary_path, "w", encoding="utf-8") as f:
        json.dump(summary, f, indent=4, ensure_ascii=False, cls=ExtendJSONEncoder)

    logger.info(f"generated task summary: {summary_path}")

'''

    project_meta = load_project_meta(test_path)
    project_root_dir = project_meta.RootDir
    conftest_path = os.path.join(project_root_dir, "conftest.py")

    test_path = os.path.abspath(test_path)
    logs_dir_path = os.path.join(project_root_dir, "logs")
    test_path_relative_path = convert_relative_project_root_dir(test_path)

    if os.path.isdir(test_path):
        file_foder_path = os.path.join(logs_dir_path, test_path_relative_path)
        dump_file_name = "all.summary.json"
    else:
        file_relative_folder_path, test_file = os.path.split(test_path_relative_path)
        file_foder_path = os.path.join(logs_dir_path, file_relative_folder_path)
        test_file_name, _ = os.path.splitext(test_file)
        dump_file_name = f"{test_file_name}.summary.json"

    summary_path = os.path.join(file_foder_path, dump_file_name)
    conftest_content = conftest_content.replace(
        "{{SUMMARY_PATH_PLACEHOLDER}}", summary_path
    )

    dir_path = os.path.dirname(conftest_path)
    if not os.path.exists(dir_path):
        os.makedirs(dir_path)

    with open(conftest_path, "w", encoding="utf-8") as f:
        f.write(conftest_content)

    logger.info("generated conftest.py to generate summary.json")


def ensure_path_sep(path: Text) -> Text:
    """ ensure compatibility with different path separators of Linux and Windows
    """
    if "/" in path:
        path = os.sep.join(path.split("/"))

    if "\\" in path:
        path = os.sep.join(path.split("\\"))

    return path



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['vhrun3',
 'vhrun3.app',
 'vhrun3.app.routers',
 'vhrun3.builtin',
 'vhrun3.ext',
 'vhrun3.ext.har2case',
 'vhrun3.ext.locust',
 'vhrun3.ext.uploader',
 'vhrun3.report',
 'vhrun3.report.allure_report',
 'vhrun3.report.html']

package_data = \
{'': ['*']}

install_requires = \
['httprunner>=3.1.6,<4.0.0']

entry_points = \
{'console_scripts': ['vhar2case = vhrun3.cli:main_har2case_alias',
                     'vhmake = vhrun3.cli:main_make_alias',
                     'vhrun = vhrun3.cli:main_hrun_alias',
                     'vhrun3 = vhrun3.cli:main',
                     'vlocusts = vhrun3.ext.locust:main_locusts']}

setup_kwargs = {
    'name': 'vhrun3',
    'version': '1.2.9',
    'description': '',
    'long_description': '\n#### 介绍\n\n基于httprunner==3.1.6版本，根据特定需求二次定制开发\n\n##### 更新日志:\n###### 1.0.1\n- 1、保留2.x版本的用例分层机制，避免冗余出现api基本信息（url、headers、method等）\n- 2、除支持http和https协议外，支持SSH协议，可以远程执行shell命令、文件上传和下载\n- 3、兼容支持2.x测试报告，便于测试时调试\n- 4、数据驱动改成一个Class N个test_*用例方式，便于用例扫描成独立用例\n- 5、支持test_xx的__doc__自动生成，并支持config.variables和parameters变量解析\n- 6、yml中config中usefixtures字段，支持pytest指定添加fixture\n###### 1.1.3\n- 7、支持特定场景下的skipif代码生成\n###### 1.1.6\n- 8、allure报告增加详细步骤信息\n###### 1.1.9\n- 9、支持testcse级别的setup和teardown\n###### 1.2.0\n- 10、Testcae和TestStep支持skipIf\n###### 1.2.1\n- 11、allure报告中显示Method和url\n- 12、修复setup_hooks/teardown_hooks使用$request和$response出现的问题\n- 13、支持api中variables与teststep中variables进行合并\n- 14、setup_hooks/teardown_hooks中返回的变量可以之后的teststep所使用\n###### 1.2.2\n- 15、setup_hooks/teardown_hooks中返回的变量显示在allure报告的extract_values中\n- 16、支持循环执行teststep直到达到目标条件或超时退出\n###### 1.2.3\n- 17、循环增强：支持loop_for和loop_while\n- 18、环境变量带上进程号+线程号，避免pytest在使用并发插件时出现混乱\n###### 1.2.4\n- 19、修复v1.2.3中的bug\n###### 1.2.5\n- 20、teradown_hooks的执行放于extract之后，以便teardown_hooks可以使用extract后的变量\n###### 1.2.6\n- 21、loop_for支持递归嵌套\n###### 1.2.7\n- 22、修复allure.step中因“{}”.format引起的错误\n###### 1.2.8\n- 23、allure中的用例名称支持变量解析\n\n#### 新增SSH部分示例：\n###### 1、执行shell命令\n```\nconfig:\n  name: demo - exec shell cmd\n  variables:\n    executor: ls\n    params: -alh\nteststeps:\n  - name: api -> shell\n    protocol: ssh\n    connection:   # 指定目标机器IP、Port、User和Password\n      ssh_ip: ${ENV(hostname)}  \n      ssh_port: ${ENV(ssh_port)}\n      ssh_user: ${ENV(ssh_user)}\n      ssh_password: ${ENV(ssh_password)}\n    request:\n      type: shell   # 指定类型为执行Shell命令\n      executor: $executor  # shell自带命令或可执行程序\n      params: $params # 字符串或字符串列表\n```\n###### 2、文件上传\n```\nconfig:\n  name: demo - upload file\nteststeps:\n  - name: api -> upload\n    protocol: ssh\n    connection:\n      ssh_ip: ${ENV(hostname)}\n      ssh_port: ${ENV(ssh_port)}\n      ssh_user: ${ENV(ssh_user)}\n      ssh_password: ${ENV(ssh_password)}\n    request:\n        type: upload # 指定类型为文件上传\n        local_path: $local_path  # 相对于本项目根目录的路径\n        remote_path: $remote_path  # 远程绝对路径\n```\n\n###### 3、文件下载\n```\nconfig:\n  name: demo - download file\nteststeps:\n  - name: api -> download\n    protocol: ssh\n    connection:\n      ssh_ip: ${ENV(hostname)}\n      ssh_port: ${ENV(ssh_port)}\n      ssh_user: ${ENV(ssh_user)}\n      ssh_password: ${ENV(ssh_password)}\n    request:\n        type: download # 指定类型为文件下载\n        local_path: $local_path  # 相对于本项目根目录的路径\n        remote_path: $remote_path  # 远程绝对路径\n```\n#### 参考：\n```\nhomepage = "https://github.com/httprunner/httprunner"\nrepository = "https://github.com/httprunner/httprunner"\ndocumentation = "https://docs.httprunner.org"\nblog = "https://debugtalk.com/\n```\n',
    'author': 'tigerjlx',
    'author_email': None,
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)
